# -*- coding: utf-8 -*-
"""
app.tasks
~~~~~~~~~

This module implements tasks to run.

:copyright: (c) 2018 European Spallation Source ERIC
:license: BSD 2-Clause, see LICENSE for more details.

"""
import os
import time
import traceback
import requests
import tower_cli
from datetime import datetime
from flask import current_app
from rq import Worker, get_current_job
from openpyxl import Workbook
from .extensions import db
from . import models, utils


class TaskWorker(Worker):
    """
    Modified version of the rq worker which updates
    the task status and end time in the CSEntry database
    """

    def save_exception(self, job, *exc_info):
        """Save the exception to the database

        The exception is only saved if it occured before the AWX job was triggered.
        If the AWX job failed, we can refer to the logs on AWX.
        """
        task = models.Task.query.get(job.id)
        if task.awx_job_id is None:
            # No AWX job was triggered. An exception occured before. Save it.
            task.exception = self._get_safe_exception_string(
                traceback.format_exception(*exc_info)
            )
            db.session.commit()

    def update_task_attributes(self, job, attributes):
        """Update the attributes of the task linked to the given job"""
        # The task is created after enqueueing the job.
        # If the job is processed very quickly, the task might
        # not be found in the database. We wait up to 3 seconds.
        for _ in range(3):
            task = models.Task.query.get(job.id)
            if task is not None:
                break
            else:
                self.log.warning("task not found...")
                time.sleep(1)
        else:
            self.log.error(
                f"Task {job.id} not found for job: {job.get_call_string()}!"
                " Task attribute not updated!"
            )
            return
        for name, value in attributes.items():
            setattr(task, name, value)
        db.session.commit()

    def update_reverse_dependencies(self, job):
        task = models.Task.query.get(job.id)
        if task is None:
            return
        task.update_reverse_dependencies()
        db.session.commit()

    def move_to_failed_queue(self, job, *exc_info):
        # This could be achieved by passing a custom exception handler
        # when initializing the worker. As we already subclass it, it's
        # easier to override the default handler in case of failure
        self.update_task_attributes(
            job, {"ended_at": job.ended_at, "status": models.JobStatus.FAILED}
        )
        self.update_reverse_dependencies(job)
        self.save_exception(job, *exc_info)
        super().move_to_failed_queue(job, *exc_info)

    def handle_job_success(self, job, queue, started_job_registry):
        self.update_task_attributes(
            job, {"ended_at": job.ended_at, "status": models.JobStatus.FINISHED}
        )
        super().handle_job_success(job, queue, started_job_registry)

    def prepare_job_execution(self, job):
        self.update_task_attributes(job, {"status": models.JobStatus.STARTED})
        super().prepare_job_execution(job)


def launch_awx_job(resource="job", **kwargs):
    """Launch an AWX job

    job_template or inventory_source shall be passed as keyword argument

    :param resource: job|workflow_job|inventory_source
    :param **kwargs: keyword arguments passed to launch the job
    :returns: A dictionary with information from resource.monitor
    """
    rq_job = get_current_job()
    job_template = kwargs.pop("job_template", None)
    inventory_source = kwargs.pop("inventory_source", None)
    if job_template is None and inventory_source is None:
        current_app.logger.warning("No job_template nor inventory_source passed!")
        return "No job_template nor inventory_source passed!"
    if job_template in (
        current_app.config["AWX_CREATE_VIOC"],
        current_app.config["AWX_CREATE_VM"],
    ) and not current_app.config.get("AWX_VM_CREATION_ENABLED", False):
        current_app.logger.info("AWX VM creation is disabled. Not sending any request.")
        return "AWX VM creation not triggered"
    if not current_app.config.get("AWX_JOB_ENABLED", False):
        current_app.logger.info("AWX job is disabled. Not sending any request.")
        return "AWX job not triggered"
    # Launch the AWX job
    tower_resource = tower_cli.get_resource(resource)
    if resource == "inventory_source":
        result = tower_resource.update(inventory_source, **kwargs)
    else:
        result = tower_resource.launch(job_template, **kwargs)
    # Save the AWX job id in the task
    task = models.Task.query.get(rq_job.id)
    task.awx_job_id = result["id"]
    db.session.commit()
    # Monitor the job until done
    result = tower_resource.monitor(pk=result["id"])
    return result


def generate_items_excel_file():
    """Export all inventory items to an excel file

    Return the name of the file
    """
    job = get_current_job()
    name = f"items-{datetime.utcnow().strftime('%Y%m%d_%H%M')}.xlsx"
    full_path = utils.unique_filename(current_app.config["CSENTRY_STATIC_FILES"] / name)
    # Instead of loading all items at once, we perform several smaller queries
    pagination = models.Item.query.order_by(models.Item.created_at).paginate(1, 100)
    wb = Workbook()
    ws = wb.active
    # Add header
    # Note that we rely on the fact that dict keep their order in Python 3.6
    # (this is official in 3.7)
    ws.append(list(pagination.items[0].to_dict().keys()))
    while pagination.items:
        for item in pagination.items:
            ws.append([val for val in item.to_row_dict().values()])
        job.meta["progress"] = int(100 * pagination.page / pagination.pages)
        current_app.logger.debug(f"progress: {job.meta['progress']}")
        job.save_meta()
        pagination = pagination.next()
    wb.save(full_path)
    return full_path.name


def create_ioc_repository(project_name):
    """Create the IOC repository on GitLab

    Return the response of the POST
    """
    api_url = current_app.config["IOC_REPOSITORY_GITLAB_URL"] + "/api/v4"
    headers = {
        "user-agent": "csentry",
        "accept": "application/json",
        "private-token": os.environ.get("GL_ACCESS_TOKEN"),
    }
    data = {
        "namespace_id": current_app.config["IOC_REPOSITORY_GROUP_ID"],
        "visibility": current_app.config["IOC_REPOSITORY_VISIBILITY"],
        "name": project_name,
        "path": project_name,
    }
    response = requests.post(api_url + "/projects", headers=headers, json=data)
    try:
        payload = response.json()
    except ValueError:
        payload = None
    if response.status_code != 201:
        current_app.logger.warning(
            f"{project_name} repository couldn't be created: {payload}"
        )
    else:
        commit_data = {
            "branch": "master",
            "commit_message": "Add .gitlab-ci.yml file",
            "actions": [
                {
                    "action": "create",
                    "file_path": ".gitlab-ci.yml",
                    "content": current_app.config["IOC_REPOSITORY_GITLAB_CI_YML"],
                }
            ],
        }
        response = requests.post(
            api_url + f"/projects/{payload['id']}/repository/commits",
            headers=headers,
            json=commit_data,
        )
    return response
